import { Lock } from "@hibas123/utils";

const ALPHABET = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ';

const ID_LENGTH = 8;

const generate = () => {
    var rtn = '';
    for (var i = 0; i < ID_LENGTH; i++) {
        rtn += ALPHABET.charAt(Math.floor(Math.random() * ALPHABET.length));
    }
    return rtn;
}
// TODO: Request Timeout
export default class Database {
    static WebSocket: typeof WebSocket;

    private initLock = new Lock();
    private socket: WebSocket;
    private requests = new Map<string, (err: Error, data: any) => void>();
    private eventhandler = new Map<string, (data: any) => void>();

    constructor(url: string, database: string, accesskey?: string, authkey?: string, rootkey?: string) {
        const u = new URL(url);
        if (authkey)
            u.searchParams.set("authkey", authkey);
        if (rootkey)
            u.searchParams.set("rootkey", rootkey);
        if (accesskey)
            u.searchParams.set("accesskey", accesskey);

        u.searchParams.set("database", database);
        this.initLock.getLock().then(lock => {
            this.socket = new Database.WebSocket(u.href);

            this.socket.onopen = () => lock.release();
            this.socket.onerror = err => { console.error(err); lock.release() };
            this.socket.onmessage = msg => {
                let message: { ns: string, data: any } = JSON.parse(msg.data as string);
                switch (message.ns) {
                    case "message":
                        this.handleResponse(message.data);
                        break;
                    case "event":
                        this.handleEvent(message.data)
                        break;
                    case "error_msg":
                        // TODO: Some kind of error handling etc.
                        console.error("Server error:", message.data);
                        break;
                }
            }
        })
    }

    private handleResponse({ id, error, data }: { id: string, error: boolean, data: string }) {
        let req = this.requests.get(id);
        if (req) {
            if (error)
                req(new Error(data), undefined);
            else
                req(undefined, data);
        } else {
            console.error("Invalid response", id);
        }
    }

    private handleEvent({ id, data }: { id: string, data: any }) {
        let eh = this.eventhandler.get(id);
        if (eh)
            eh(data);
    }

    private async queryRequest(type: "get" | "set" | "push" | "keys" | "subscribe" | "unsubscribe", path: string[] | string, data?: any, id?: string) {
        await this.initLock.getLock().then(l => l.release());
        if (typeof path === "string")
            path = path.split("/").filter(e => e !== "");
        if (!id)
            id = generate();

        return new Promise<any>((yes, no) => {
            this.requests.set(id, (err, data) => {
                err ? no(err) : yes(data)
                this.requests.delete(id);
            });
            this.socket.send(JSON.stringify({ ns: "query", data: { id, type, path, data } }))
        })
    }

    keys(path: string | string[]) {
        return this.queryRequest("keys", path);
    }

    get<T = any>(path: string | string[]): Promise<T> {
        return this.queryRequest("get", path);
    }

    set(path: string | string[], data: any) {
        return this.queryRequest("set", path, data);
    }

    push(path: string | string[], data: any): Promise<string> {
        return this.queryRequest("push", path, data);
    }

    async subscribe(path: string | string[], type: "push" | "set", ondata: (data: any) => void) {
        const id = generate();
        const subscriptionID = await this.queryRequest("subscribe", path, type, id);
        this.eventhandler.set(subscriptionID, ondata);
        return {
            unsubscribe: () => {
                this.queryRequest("unsubscribe", path, undefined, id);
                this.eventhandler.delete(subscriptionID);
            }
        }
    }
}